//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
//
// @author Brian Ricks
//

#include "RadioPacketGenerator.h"

#include <omnetpp.h>

Define_Module(RadioPacketGenerator);

RadioPacketGenerator::RadioPacketGenerator() {
    nb = NULL;
    isScheduled = false;
}

RadioPacketGenerator::~RadioPacketGenerator() {
    // TODO Auto-generated destructor stub
}

void RadioPacketGenerator::initialize(int stage) {
    if (stage == 0) {
        nb = NotificationBoardAccess().getIfExists();
        generatedPktLength = par("packetLength");
        sendDelay = par("packetSendDelay");

        if (!nb) {
            error("NotificationBoard reference is NULL!");
        }

        // subscribe to changes in radio state (note this must be done in stage 0 as
        // the initial state notifications are fired in stage 1).
        nb->subscribe(this, NF_RADIOSTATE_CHANGED);

        _radioModuleID = gate("radioOut")->getNextGate()->getOwnerModule()->getId();  // Get the module ID of the radio

        scheduleMsg = new cMessage("scheduleMsg");
    }
}

// Processes self messages as they are received (think of these as timers)
// Any other messages received (such as those received by the radio when it isn't transmitting)
// are discarded.
void RadioPacketGenerator::handleMessage(cMessage *msg) {

    if (msg == scheduleMsg) {
        // Time to send a message!!
        if (_radioState == RadioState::TRANSMIT) {
            // wait, what?  Well we can't send now!
            // This will happen if the time it takes to transmit a packet is greater than the send delay.
            // Wait and send when the state becomes desirable.
            isScheduled = false;

            return;
        }

        sendPacket();
        //schedule the next one
        scheduleAt(simTime() + sendDelay, scheduleMsg);
    }
    else {
        // nothing for us, delete
        EV << "[JAMMER]: Received message from Radio, discarding...\n";

        delete msg;
    }
}

// Handler for notifications that we are subscribed to (radio state changes)
void RadioPacketGenerator::receiveChangeNotification(int category, const cObject *details) {

    Enter_Method_Silent(); // required as this method is callable from other modules (NotificationBoard)

    if (category == NF_RADIOSTATE_CHANGED) {
        // this is the only category we formally subscribed to
        const RadioState * radioState = check_and_cast<const RadioState *>(details);

        // Check which radio module sent the event.  We only want the radio module we are sending packets to.
        if (radioState->getRadioId() != _radioModuleID) {
            return;
        }

        _radioState = radioState->getState();

        if (isScheduled || _radioState == RadioState::TRANSMIT) {
            // The radio is busy (probably sending our last packet), so ignore this notification
            // Note that in radio.cc, the only time a packet won't be transmitted is if the radio state is TRANSMIT
            // If there is a sendDelay specified, then the isScheduled flag will always be set unless a scheduled
            // packet couldn't be transmitted, or if the simulation just started (first packet sent out).
            return;
        }

        // Either there is no sendDelay, or we couldn't schedule a new packet to be sent.  Send one now.
        sendPacket();

        if (sendDelay > 0) {
            // This means a packet was not scheduled earlier due to the radio being busy
            // (usually a sign that the sendDelay < packet transmit time), or if the first packet was just sent out.
            scheduleAt(simTime() + sendDelay, scheduleMsg);
            isScheduled = true;
        }
    }
}

void RadioPacketGenerator::sendPacket() {
    // Time to send a packet!!
    cPacket *newPkt = new cPacket("RadioPacketGeneratorPkt");
    // set the length of the packet
    newPkt->setByteLength(generatedPktLength);

    send(newPkt, "radioOut");

    EV << "[JAMMER]: Delivered a packet to the radio for transmission\n";
}

void RadioPacketGenerator::finish() {
    cancelAndDelete(scheduleMsg);
}
